; This is the game driver for the VGA defender game. (c) Design-Design Software

; Play a game, given al = nz for replay

Play_Game:	mov cs:Key_Replay_F,al	; Set the key mode

		; First, start from scratch

		mov al,cs:optMen	; Population
		mov ah,0		;
		mov cs:nMen_Coll,ax	;

		mov w cs:Score,0	; No score yet

		mov b cs:Shield_D,63	; Start with max shields

		mov w cs:Magnum_Shots,0	; None yet

		; Setup options

		call Calc_Options	; Make them valid

		; Get the game keyboard

		mov al,cs:fForce_AT	; Force to AT keys
		call Link_Kbd		; Enable the game keys

		; Assume full screen

		mov b cs:Minature_F,0	; Set big

		; In replay mode ?

		test b cs:Key_Replay_F,-1 ; Replay ?
		jz PG_Size		  ; Skip

		mov b cs:Minature_F,1	; Set small

		; Now, setup the screen-size

PG_Size:	test b cs:Minature_F,-1	; Are we in the large screen mode
		jnz PGS_1		; No, skip

		; Reduce the screen size

		call Set_Full_Screen	; Set the full screen

		jmp s PGS_2		; Skip

	; Small window mode

PGS_1:		call Set_Small_Window	; Set the window mode

		; Draw the frame

PGS_2:		call Draw_Frame		; Show it

		; Duplicate pages

		call Duplicate_Page	; Copy it

		; Now, point at the sprites

		call Select_Sprites	; Select the large/small set

		; Kill cheating

		mov b cs:Cheat_F,0	; Not cheating yet

		; Setup the first sheet

		mov al,cs:Initial_Level	; Get the starting sheet number
		mov cs:Sheet_No,al	; The first sheet

		; Play it
 
Game_Loop:	call Play_Sheet		; Play it

		mov al,cs:Game_Flag	; Why exit ?

		cmp al,1		; Resigned ?
		je Game_Ends		; Yes, exit to front-end

		cmp al,2		; Killed them all ?
		je GL_Next_Sheet	; Yes, skip

		; Killed, explode and return

		jmp Game_Ends		; Failed, try again

	; Start the next sheet

GL_Next_Sheet:	mov al,cs:Sheet_No	; Get the current number

		cmp al,nSheets		; Done ?
		je Game_Ends		; Yes, exit

		; Any men left ?

		test w cs:nMen_Coll,-1	; Any ?
		je Game_Ends		; No, exit

		; Step the sheet number

		inc al			; + 1
		mov cs:Sheet_No,al	; Save it

		jmp s Game_Loop		; Loop

	; Exit to front-end

Game_Ends:	call Unlink_Kbd		; Release the keys

		; Set the page

		mov b cs:Front_Screen,0	; Front

		mov ax,cs:Score		; Score

		ret			; Leave the game

; Set the full-screen mode

Set_Full_Screen:mov w cs:VDU_Offset,0	; Offset = 0

		ret			; Done

; Set the window mode

Set_Small_Window: mov ax,(Y_Step_640/4) + 60 * Y_Step_640 ; Middle
		mov w cs:VDU_Offset,ax ; Offset to window

		ret			; Done

; Wait for the fire key

Wait_Fire:	test b cs:Key_Replay_F,-1 ; Are we replaying ?
		jnz WF_Pause		  ; Yes, pause

		call WF_Pressed		; Wait while pressed
		call WF_NotPressed	; Wait while not pressed

		; Now, do we have the cheat key pressed ?

		test b cs:Key_Matrix+2Eh,1 ; Cheat ?
		jz WF_Pressed		   ; No, skip

		; Yes, set the cheat-mode

		mov b cs:Cheat_F,1	; Assert cheat-mode

		; Fall into

WF_Pressed:	test b cs:Key_Matrix+1,1; Esc ?
		jnz WF_Exit		; Yes

		mov di,o Key_Fire	; Test the fire key
		mov cx,kfFire		;
		mov ax,0		;

		call GKS_UDK		;

		cmp ax,0		; Pressed ?
		jnz WF_Pressed		; Yes, loop

WF_Exit:	ret			; Done

WF_NotPressed:	test b cs:Key_Matrix+1,1; Esc ?
		jnz WF_Exit		; Yes

		mov di,o Key_Fire	; Test the fire key
		mov cx,kfFire		;
		mov ax,0		;

		call GKS_UDK		;

		cmp ax,0		; Pressed ?
		jz WF_NotPressed	; No, loop

		ret			; Done

	; Small delay

WF_Pause:	mov cx,60		; About a second

WFP_Lp:		push cx			; Save
		call Next_Frame		; Wait a bit
		pop cx			; Restore

		loop WFP_Lp		; Loop

		ret			; Done

; Given al=sheet number, setup for this sheet

Setup_Sheet:	mov bx,cs:nMen_Coll	; Only the saved ones
		mov dx,bx		; Add 1/2 more
		sar dx,1		; 1/2
		add bx,dx		;		
		mov cs:nMen,bx		;

		; Other gumph

SS_Vars:	mov ah,0		; ax = 1 .. n
		dec ax			; 0 .. n-1

		mov dx,lSheet_Defn	; Length of data

		mul dx			; bx := index into table
		mov bx,ax		;

		add bx,o Sheet_Defns	; Point at the data

		mov ax,DATA		; Setup the segment
		mov ds,ax		;

		; Now, get the data for this sheet

		mov al,dsMount_Col[bx]	; Get the colour
		mov cs:Mount_Colour,al	; Set it

		; Point at the name

		mov ax,dspName[bx]	; Get the name
		mov ds:pName,ax		; into DS

		; Point at the attack-wave

		mov ax,dspWave[bx]	; Get the wave-info
		mov ds:pWave,ax		; into DS

		mov w cs:Wave_Delay,1	; Delay to next-wave

		; Done, exit

		ret			; Done

; Get the next creature for this attack-wave

Wave_Get_Next:	mov ax,DATA		; Use DS
		mov ds,ax		;

		mov si,ds:pWave		; Point at the next one

		; Get this entry

		mov bx,ds:dwCode[si]	; Point at the routine
		mov ax,ds:dwTime[si]	; Delay to next
		mov cx,ds:dwNext[si]	; Point at the next

		lea si,6[si]		; Skip these

		cmp cx,0		; Valid ptr ?
		jz WGN_1		; No, skip

		mov si,cx		; Yes, use it

		; Save params

WGN_1:		test b cs:Difficulty,02h; Slower ?
		jnz WGN_2		; No, skip

		mov cx,ax		; 1/2 more delay
		shr cx,1		;
		add ax,cx		;

WGN_2:		mov cs:Wave_Delay,ax	; Save the delay
		mov ds:pWave,si		; Save to ptr

		; Add this (these) creatures

		call bx			; Call the code

		; Done

		ret			; Done, exit

; Clear either the 320 or the 640 display

CLS_Size:	test b cs:Minature_F,-1	; Small ?
		jz vCLS			; No, big

		jmp CLS_320		; Clear the small window

vCLS:		jmp CLS			; Clear the screen

; Show the pause display

Pause_Screen:	call CLS_Size		; Clear it

		; Show the title

		mov si,o sPause1	; Planet title (big)

		test b cs:Minature_F,-1	; Small ?
		jz PSa			; No, skip

		mov si,o s_S_Pause1	; Small title

PSa:		call Print_Str		; Show it

		; Show the planet's name

		mov ax,DATA		; Seg
		mov ds,ax		;

		mov si,ds:pName		; Get the name
		call Print_Str_DS	; Show it
		
		; End it

		mov si,o sPause2	; Big title

		test b cs:Minature_F,-1	; Small ?
		jz PSb			; No, skip

		mov si,o s_S_Pause2	; Small title

PSb:		call Print_Str		;

		; Pause

		call Wait_Fire		; Delay

		; Done

		jmp CLS_320		; Clear it

; Play one sheet

Play_Sheet:	mov cs:PlayStack,sp	; Save the stack

		; Setup vars

		call Init_Sheet		; Start the sheet

		; First undraw everything

Sheet_Loop:	call Undraw_Old		; Remove all the shit

		; No action yet

		mov b cs:Action,0	; Nothing happened yet

		; Now, get the key-states

		call Get_Key_States	; Get them

		; Is the bomb key pressed ?

		mov ax,cs:Key_Flags	; Is the smart key pressed ?
		and ax,kfSmart		; Get the flag
		or al,ah		; Get flag into al

		mov ah,0		; Kill smart bomb

		cmp cs:Smart_Key_P,al	; Same
		jz SL_NSB		; Yes, skip

		mov cs:Smart_Key_P,al	; Save the new state
		cmp al,0		; Pressed ?
		je SL_NSB		; No, skip

		; The smart-bomb key is pressed, drive penalty

		mov al,cs:optSmart_Rate	; The damage
		call Shield_Dec		;

		mov ah,1		; Set bomb

		; Now set the bomb

SL_NSB:		mov cs:Smart_Bomb_F,ah	; Drive the flag

		; Now move the ship

		call Move_Ship		; Drive the player about

		; Now calc the new window posn.

		call Move_Window	; Drive the viewport about

		; Now we draw the new mountains

		mov ax,cs:Scr_X		; Old posn

		call Draw_Mount		; Remove them

		; Setup for the undraw

		call Set_Undraw_Sp	; Start the sprite stack

		; Show the things

		call Drive_Things	; Show them

		; Show the lasers

		call Drive_Lasers	; Show them

		; Show the projectiles

		call Drive_Projs	; Show them

		; If there are frags, we are in action mode 1

		test w cs:pFrags,-1	; Any frags ?
		jz PL_1			; No, skip

		; We are in mode 1

		mov b cs:Action,1	; Mode 1

		; Show the ship in the new posn

PL_1:		call Draw_Ship		; Show it

		; Now the action-replay marker

		mov al,13		; Replay sprite

		test b cs:Key_Replay_F,-1 ; Are we replaying ?
		jnz PL_Replay		; Yes, skip

		; Now the cheat marker

		mov al,14		; Cheat sprite

		test b cs:Cheat_F,-1	; Are we cheating ?
		jnz PL_Replay		; Yes, skip

		; Finished drawing sprites

PL_NR:		call Finish_Undraw	; Done

		; Show the fragments

		call Drive_Frags	; Show them

		; Show the score, clock etc

		call Show_Display	; Show it

		; Store the undraw list

		call Draw_End		; Save it

		; Release the landing pads (if unoccupied)

		call Release_Pads	; Dec the occupied counters

		; Did we do anything ?

		test b cs:Action,-1	; Any action ?
		jz Play_None		; No, skip

		; Escape ?

		test w cs:Key_Flags,kfEsc ; Esc ?
		jnz Play_Esc		  ; Yes, done

		; Have we finished ?

		mov al,cs:Game_Flag	; Done ?
		cmp al,0		;
		jnz Play_Exit		; Yes, skip

		; Now, is it time for the next monster ?

		dec w cs:Wave_Delay	; Dec pause
		jnz Play_Wait		; Not yet, skip

		call Wave_Get_Next	; Giz the next thing

		; Wait for the next frame

Play_Wait:	inc b cs:Frame_Cnt	; Step the count

		; Paused ?

		mov di,o Key_Pause	; Pause key
		call Test_UDK		;
		jnz Pause_Game		; Yes, skip
		
		; Boss ?

		mov di,o Key_Boss	; Boss key
		call Test_UDK		;
		jnz vBoss_Exit		; Shit, exit
		
		call Game_Next_Frame	; Wait till sync'd

		; Set the vdu page, etc

		call Swap_Pages		; Swap them

		; Loop

		jmp Sheet_Loop		; Loop

	; Show the action-replay sign

PL_Replay:	mov bl,cs:Frame_Cnt	; Get the count
		and bl,020h		; On ?
		jz PLR_1		; No, skip

		mov bl,0Eh		; Bright yellow

PLR_1:		mov dx,5		; X
		mov ah,Min_Y+5		; Y

		mov bh,0Ch		; Red surround

		call Draw_Sprite	; Draw it

		jmp s PL_NR		; Back again

	; We have cleared everything

Play_None:	mov al,2		; Cleared them
		jmp s Set_GO		; Flag it		

; Pause key

Pause_Game:	call Pause_Key		; Get flag
		jnz Pause_Game		; Wait until release

PG_1:		call Pause_Key		; Wait until press
		jz PG_1			;

PG_2:		call Pause_Key		; Wait until release
		jnz PG_2		;

		; Continue

		call Swap_Pages		; Swap them

		jmp Sheet_Loop		; Loop

;

vBoss_Exit:	jmp s Boss_Exit		; Vector

	; Scan the pause-key, NZ = pressed

Pause_Key:	sti			; EI

		mov di,o Key_Pause	; Pause key
		jmp Test_UDK		;

	; Assert ESC

Play_Esc:	call PE_Wait		; Wait until ESC free

		; Are we cheating ?

		mov al,2		; Cleared sheet
		mov w cs:nMen_Coll,5	; Assume 10 men

		test b cs:Cheat_F,-1	; Cheat mode ?
		jnz Set_GO		; Yes, skip

		; flag it

		mov al,1		; Resigned

Set_GO:		mov cs:Game_Flag,al	;

		; Done, exit

Play_Exit:	call End_Keys		; Add a fake escape here 

		mov al,cs:Game_Flag	; Killed ?

		cmp al,1		; Escape ?
		je Big_Bang		; Yes, skip

		cmp al,3		;
		jnz Play_Exit1		; No, skip

Big_Bang:	call Explode		; Explode the ship

Play_Exit1:	call Medium_Delay	; Pause

		; Restore the video page

		mov al,0		; Page = 0
		call Set_VDU_Page	;
		call Set_DRAW_Page	;

		; Set the display mode, if we are not in the small screen

		test cs:Minature_F,-1	; Small ?
		jnz PEa			; Yes, skip

		call Set_640_350	; Set the video mode

		call Small_Delay	; Wait a bit

		; Exit

PEa:		mov sp,cs:PlayStack	; Restore

		ret			; Done

; Loop until the escape key is released

PE_Wait:	sti			; EI

		test b cs:Key_Matrix+1,1; Esc ?
		jnz PE_Wait		; Yes, wait

		ret			; Done

	; We have the boss key, exit

Boss_Exit:	mov b cs:Game_Flag,1	; Resigned

		; Done, exit

		call End_Keys		; Add a fake escape here 

		; Restore the video page

		mov al,0		; Page = 0
		call Set_VDU_Page	;
		call Set_DRAW_Page	;

		; Now, setup the text display

		mov al,03H		; Else set it to 80x25 colour

		mov ah,0		;
		int vdu			;

		; Cursor off

		mov ch,20h		; Start line
		mov cl,20h		; End line

		mov ah,1		; Set size
		int vdu			; Call it

		; Draw the display

		mov ax,DATA		; Source
		mov ds,ax		;

		mov si,o Boss_Screen	; 

		mov ax,0B800h		; Dest
		mov es,ax		;

		mov di,0		; 

		mov cx,80*25		; Char screen
	rep	movsw			; Copy it

		; Wait for an escape

		; Escape ?

Boss_Wait:	call Next_Frame		; Pause here

		test b cs:Key_Matrix+1,1; Esc ?
		jz Boss_Wait		; No, loop

		; Exit

		call Set_640_350	; Set the video mode

		; Exit

		jmp s PEa		; Exit

; Add a number to the score

Add_Score:	add cs:Score,ax		; Add ax to it
		ret			; Done

; Wait for the display to settle

Small_Delay:	mov cx,10		; 10 frames
		jmp s SD_Lp		;

Medium_Delay:	mov cx,50		; 50 frames
		jmp s SD_Lp		;

SD_Lp:		call Next_Frame		; Wait one frame
		loop SD_Lp		; Loop

		ret			; Done

; Wait in the game

Game_Next_Frame:test b cs:Language,03h	; Assembler ?
		je vNext_Frame		; Yes, skip

		; Basic ?

		cmp b cs:Language,3	; Basic ?
		je GNF_Basic		; Yes, skip

		cmp b cs:Language,2	; C++ ?
		je GNF_CPlus		; Yes, skip

		; C

		mov cx,2		; Sloow
		jmp s SD_Lp		;

	; Emulate basic

GNF_Basic:	mov cx,16		; SSSLLLOOOWWWLLLYYY
		jmp s SD_Lp		; As above
		
	; Emulate C++

GNF_CPlus:	call Rand_32		; Random
		and ax,0003h		; 0 .. 3
		inc ax			; 1 .. 4
		mov cx,ax		; SSSLLLOOOWWWLLLYYY
		jmp s SD_Lp		; As above
		
        ; Jump

vNext_Frame:	jmp Next_Frame		; Wait for a frame

; Explode the ship

Explode:	mov b cs:Explode_F,1	; Set the flag

		; Now, setup the explosion

		call Explode_Ship	; Blow up the ship

		; Wait a while

		mov cx,4		; Slight pause

Exp_Pause:	push cx			; Save
		call Next_Frame		; Pause
		pop cx			; Restore

		loop Exp_Pause		; Loop

		; Show it

Explode_Loop:	call Undraw_Old		; Remove all the shit

		; Now we draw the new mountains

		mov ax,cs:Scr_X		; Old posn

		call Draw_Mount		; Remove them

		; Setup for the undraw

		call Set_Undraw_Sp	; Start the sprite stack

		; Show the things

		call Drive_Things	; Show them

		; Show the projectiles

		call Drive_Projs	; Explode them

		; Now the action-replay marker

		mov al,13		; Replay sprite

		test b cs:Key_Replay_F,-1 ; Are we replaying ?
		jnz EX_Replay		  ; Yes, skip

		; Now the cheat marker

		mov al,14		; Cheat sprite

		test b cs:Cheat_F,-1	; Are we cheating ?
		jnz EX_Replay		; Yes, skip

		; Finished drawing sprites

EX_NR:		call Finish_Undraw	; Done

		; Show the fragments

		call Drive_Frags	; Show them

		; Show the score, clock etc

		call Show_Display	; Show it

		; Store the undraw list

		call Draw_End		; Save it

		; Wait for the next frame

		inc b cs:Frame_Cnt	; Step the count

		call Game_Next_Frame	; Wait till sync'd

		; Set the vdu page, etc

		call Swap_Pages		; Swap them

		; Escape ?

		test b cs:Key_Matrix+1,1; Esc ?
		jnz Exp_Esc		; Yes, done

		; Loop

		test w cs:pProjs,-1	; Any left ?
		jnz Explode_Loop	; Loop

;		test w cs:pFrags,-1	; Any left ?
;		jnz Explode_Loop	; Loop

		; Escape

Exp_Esc:	call Undraw_Old		; Clear them
		call Swap_Pages		; Show it

		call PE_Wait		; Wait for the ESC key to be released

		ret			; Done

	; Show the action-replay sign

EX_Replay:	mov bl,cs:Frame_Cnt	; Get the count
		and bl,020h		; On ?
		jz EXR_1		; No, skip

		mov bl,0Eh		; Bright yellow

EXR_1:		mov dx,5		; X
		mov ah,Min_Y+5		; Y

		mov bh,0Ch		; Red surround

		call Draw_Sprite	; Draw it

		jmp s EX_NR		; Back again

; Setup the start of game vars

Init_First:	call Init_Ship		; Setup the ship's vars

		mov w cs:Scr_X,Total_Width-offL;0	; Start at 0

		mov w cs:Ship_X,0;offL	; On screen
		mov b cs:Ship_Y,100	;

		mov w cs:Wind_Offset,0;offL; Window offset

		mov b cs:Ship_Dir,01h	; Right

		mov b cs:Ship_Speed,0	; Starting speed

		mov b cs:Ship_S_Del,0	; Clear the delay

		mov b cs:Thrust_F,0	; No flames

		mov b cs:Frame_Cnt,0	; Count := 0

		; Start the undraw stack

		jmp Init_Pages		; Setup the stack of items to undraw

; Setup the next sheet vars

Init_Sheet:	call Init_Keys		; Start things here

		; Setup this sheet

		mov al,cs:Sheet_No	; The current sheet
		call Setup_Sheet	; Set it up

		; Show info

		call Pause_Screen	; Info page

		; Now, is this a minature display ?

		test b cs:Minature_F,-1	; Mini screen ?
		jnz IS_VDU_Setup	; Yes, skip

		; Setup the display

		call Set_320_200	; Set the video mode

		call Small_Delay	; Wait a bit

		; Setup VDU_Seg

		mov ax,0		; Offset = 0

IS_VDU_Setup:	mov ax,cs:FRONT_Seg	; Start of display
		mov cs:VDU_Seg,ax	; Set it

		; Setup generic vars

		mov b cs:Game_Flag,0	; Not over yet

		mov cs:nMen_Coll,0	; None yet

		call Init_First		; Setup for the start

		; Setup the Radar divide

		call Init_Radar		; Setup the divide look-up

		; Setup the free-space list

		call Init_Free		; Create the free space list

		; Setup the display

		call Init_Display	; Setup the display

		; Setup the explosion fragments

		call Init_Frags		;

		; Setup the laser bolts

		call Init_Lasers	;

		; Setup the projectiles

		call Init_Projs		;

		; Now, setup the things

		call Init_Things	; Setup the monsters
		call Create_Things	; 

		; Smart bomb timer

		mov b cs:Smart_Key_P,1	; Pressed

		; Clear the number of bombs allowed

		mov w cs:Bomb_Count,0	; None yet

		; Not exploding

		mov b cs:Explode_F,0	; Clear the flag

		; Build the mountains

		jmp Mount_Build		; Build them

; Extra code

		include drawing.inc	; Code to draw "sprites"

		include movers.inc	; Code to move "sprites"

		include things.inc	; Aliens

		include top.inc		; Display driving code

		include keys.inc	; Keyboard drivers

; Vars
		align 2			; Even

Wave_Delay	dw ?			; Frames to next attack-wave

PlayStack	dw ?			; Temp stack hold

Scr_X		dw ?			; LHS of screen

Ship_X		dw ?			; Ship's X
Ship_Y		db ?			;	 Y

Ship_Dir	db ?			; 1 = right

Ship_Speed	db ?			; Speed of ship

Ship_S_Del	db ?			; Speed change delay

Thrust_F	db ?			; Thrust flag

Game_Flag	db ?			; NZ = type of end

Action		db ?			; NZ = something happened

Frame_Cnt	db ?			; Increments at frame rate

Sheet_No	db ?			; First = 1

Smart_Bomb_F	db ?			; Kill visible things
Smart_Key_P	db ?			; Rate of loss

Explode_F	db ?			; NZ = exploding

Minature_F	db ?			; NZ = small window

Cheat_F		db ?			; NZ = cheating
