; Drive the projectiles

Proj_Range	equ 320			; One screen

; Right, fire something at the ship

Fire_At_Ship:	mov dx,ax		; Save

		; Is this in range ?

		mov bx,cs:Ship_X	; Target
		mov ax,cs:d_X[di]	; Our X
		call Calc_Offset	; Vector
		mov cs:Temp_dX,ax	; Save		

		; Is this in range

		cmp ax,0		; +/-
		jns FAS_Pve		; Skip

		neg ax			; Abs

FAS_Pve:	cmp ax,Proj_Range	; In range ?
		jb FAS_Fire		; Yes, try it

FAS_Explode:	ret			; Done, out of range

	; Try to fire at the ship

vFAS_X:		jmp FAS_X		; Vector

FAS_Fire:	test b cs:Explode_F,-1	; Are we allowed to fire ?
		jnz FAS_Explode		; No, exit

		mov si,di		; Use si to point at the thing

		; Get some free space

		call Get_Free		; Find a free space
		jc vFAS_X		; Can't, exit

		; Set it to the ship's position

		; X

		mov ax,cs:d_X[si]	; X
		mov cs:d_X[di],ax	;

		; Y

		mov al,cs:d_Y[si]	; Y
		mov cs:d_Y[di],al	;

		; Now, calculate the vector to the ship

		; X

		mov ax,cs:Temp_dX	; Get it
		mov cx,ax		; Save
		mov cs:ddX[di],ax	; temp

		; Y

		mov al,cs:Ship_Y	; Target
		sub al,cs:d_Y[di]	; Our 
		cbw			; 16-bit
		mov bx,ax		; Save dY

		; Now, we must scale these to get the required speed.

		cmp cx,0		; dX +ve ?
		jns FAS_ND_1		; Yes, skip

		neg cx			; Abs dX

FAS_ND_1:	cmp ax,0		; dY +ve ?
		jns FAS_ND_2		; Yes, skip

		neg ax			; Abs dX

FAS_ND_2:	add cx,ax		; |dx| + |dY|
		inc cx			; NOT 0 !

		; Calc the scale speed

		test b cs:Difficulty,02h; Slower ?
		jnz FAS_Diff		; No, skip

		shr dx,1		; /2 (half speed)

FAS_Diff:	mov ax,dx		; Speed
		mov dx,0		; 32-bit

		div cx			; Calc 

		; Now scale the real dX/dY

		mov cx,ax		; Scale factor
		inc cx			; NOT ZERO !

		mov ax,cs:ddX[di]	; dX
		imul cx			; Scale it
		mov cs:ddX[di],ax	;

		mov ax,bx		; dY
		imul cx			; Scale it
		mov cs:ddY[di],ax	;

		; Shape

		mov b cs:dShape[di],32	; Proj shape

		; Alive

		mov b cs:dFlags[di],01h	; Make it live, aimed at things

		; Setup colours

		mov w cs:dCol1[di],00E0Ch ; Col 0,1

		; Setup life

		mov al,cs:optProj_Life	; Lifetime
		mov ah,0		;

		add ax,ax		; * 4
		add ax,ax		;

		mov cs:dLife[di],ax	;

		; Add it to the list

		call Add_Proj		; Add it to the list

		; Done, exit

FAS_X:		mov di,si		; Restore ptr
		ret			;

; Show the projectiles

Drive_Projs:	mov di,cs:pProjs	; Point at the Proj bolt list

		mov ax,o pProjs-dpNext	; Point at this ptr
		mov cs:pPrev_Proj,ax	; Save this

		cmp di,0		; Any things ?
		jz DP_X			; No, exit

		; Explosion noise ?

		test b cs:Explode_F,-1	; Exploding ?
		jz DP_Quiet		; No, skip

		; Start the noise

		mov ax,2000		; Different
		call Set_Freq		;

		call Spk_On		; Sound On

		; There are Projs

DP_Quiet:	or b cs:Action,2	; Flag that we did something

DP_Lp:		cmp di,0		; Valid ?
		jz DP_X			; No, skip

		; Dec it's life-time

		dec w cs:dLife[di]	; Decay it
		jz DP_End_TO		; Kill it

		; Now show this projectile

		call Move_Proj		; Move it
		jc DP_End		; Kill it, (explode)

		; Move onto the next

DP_Next:	mov cs:pPrev_Proj,di	; Save a ptr to it

		mov di,cs:dpNext[di]	; Next Proj

		jmp DP_Lp		; Loop

	; Exit code, setup the un-draw operation

DP_X:		call Spk_Off		; Sound off

		ret			; Done

	; We must kill this Projectile

DP_End:		mov ax,cs:dpNext[di]	; The next one
		mov bx,cs:pPrev_Proj	; The last one

		mov cs:dpNext[bx],ax	; The last points at the next

		; Add this one to free-space

		call Put_Free		; Release di

		; Now, drive the next one

		mov di,ax		; Point at the next one

		jmp s DP_Lp		; Loop

	; Time it out

DP_End_TO:	mov ax,cs:dpNext[di]	; The next one
		mov bx,cs:pPrev_Proj	; The last one

		mov cs:dpNext[bx],ax	; The last points at the next

		; Add this one to free-space

		call Put_Free		; Release di

		; Explode it

		test b cs:Explode_F,-1	; Exploding the ship ?
		jnz DP_NoFrag		; Yes, skip

		push ax			; Save

		mov ax,cs:d_X[di]	; X
		mov dl,cs:d_Y[di]	; Y

		mov bx,o Frag_Explode_SS; Very small table
		call Gen_Frag		; Generate the fragments

		pop ax

		; Now, drive the next one

DP_NoFrag:	mov di,ax		; Point at the next one

		jmp s DP_Lp		; Loop

; This is an explosion fragment, allow for gravity

MP_Expl_Frag:	mov ax,cs:ddY[di]	; Gravity
		add al,cs:optExpl_Grav	;
		adc ah,0		;
		mov cs:ddY[di],ax	;

		; Time to change ?

		dec b cs:dTemp1[di]	; Time yet ?
		jnz MP_NotHit		; No, skip

		; Yes, get the new time, shape and colour

		mov bx,cs:dTemp2[di]	; Point at it

		mov ax,cs:[bx]		; Time/Shape

		mov cs:dTemp1[di],al	; Time
		mov cs:dShape[di],ah	; Shape

		inc bx			; Skip
		inc bx			;

		mov cs:dTemp2[di],bx	; Save the ptr

		; Done

		jmp s MP_NotHit		; Move as above

; Entry with di pointing at data

Move_Proj:	test b cs:dFlags[di],080h ; Explosion ?
		jnz MP_Expl_Frag	  ; Yes, skip

		mov al,cs:Ship_Y	; The ship's Y
		sub al,3		; Widen it
		cmp al,cs:d_Y[di]	; Get the highest possible Y
		jb MP_PossHit		; Yes, possible hit

		; Move it

MP_NotHit:	mov bl,cs:ddX[di]	; Step in X
		mov al,cs:ddX+1[di]	;
		cbw			;

		add bl,cs:d_frac_X[di]	; Add it
		adc ax,cs:d_X[di]	;

		mov cs:d_frac_X[di],bl	; Save fractional X

		call Wrap_X		; Clip it

		mov cs:d_X[di],ax	; Save new X

		; Move it in Y

		mov ax,cs:ddY[di]	; Disp
		add ax,cs:d_frac_Y[di]	; Posn

		mov cs:d_frac_Y[di],al	; Save fraction

		mov al,ah		; Y

		; Clip Y

		cmp al,Max_Y-3		; Below bottom ?
		ja MP_Kill		; Above

		cmp al,Min_Y+6		; Above top ?
		jb MP_Kill		; No, skip

		mov cs:d_Y[di],al	; Save pos

		; Show this proj.

MP_X:		call Draw_Thing		; Show it

		clc			; Ok
		ret			; Done

MP_Kill:	stc			; Kill it
		ret			;

; Have we hit the ship ?

MP_PossHit:	add al,3+3		; Down
		cmp al,cs:d_Y[di]	; Get the highest possible Y
		jb MP_NotHit		; No, not hit

		; The Y's are right, check the X's

		mov ax,cs:Ship_X	; The ship's position

		sub ax,16		; Back 11
		call Wrap_X		;
		mov cx,ax		;

		mov ax,cs:Ship_X	; The ship's position

		add ax,16		; Back 11
		call Wrap_X		;
		mov dx,ax		;

		mov ax,cs:d_X[di]	; X in range ?

		cmp cx,dx		; Wrapped ?
		jnc MP_CheckEnd		; Yes, try another clip

		; Now try the X's

		cmp ax,cx		;
		jb MP_NotHit		; No, fail

		cmp ax,dx		; X in range ?
		ja MP_NotHit		; No, fail

		; Assume a hit

MP_Hit:		mov al,cs:Sheet_No	; Dec the shields
		shr al,1		; / 2
		inc al			; + 1
		call Shield_Dec		;

		; Explode it

		mov ax,cs:d_X[di]	; X
		mov dl,cs:d_Y[di]	; Y

		mov bx,o Frag_Explode_SS; Very small table
		call Gen_Frag		; Generate the fragments

		; Kill this projectile

		jmp s MP_Kill		; Kill the proj

	; The proj is at the end, alternate clip

MP_CheckEnd:	cmp ax,dx		; X<RHS
		jb MP_Hit		; Yes, skip

		cmp ax,cx		; LHS<X
		ja MP_Hit		; Yes, skip 

		; Not hit, skip

		jmp MP_NotHit		; Skip

; Vars

		align 2			; Even

Temp_dX		dw 0			; Temporary hold for dX
pPrev_Proj	dw 0			; Ptr to the last ptr