fWidth		equ 15			; Fifteen cells wide
fDepth		equ 23			; 23 Cells deep

Initial_Speed	equ 32*15		; Time per step (15 frames)
Loop_Speed	equ 32*8		; Medium 	(8 frames)
Base_Speed	equ 32*4		; Fastest	(4 frames)
Step_One	equ -1			; Dropped time
Step_Two	equ -4			; Not dropped time

Wide		equ 80			; Screen width

tkLeft		equ 1			; Dummy key values
tkRight		equ 2			;
tkRotL		equ 3			;
tkRotR		equ 4			;
tkDrop		equ 5			;
tkEsc		equ 6			;

kRep_Cnt	equ 1			; 1/70 of a sec.

; This is a simple "TETRIS" like game as a diversion

Play_Tetrus:	mov al,03H		; Else set it to 80x25 colour

		mov ah,0		;
		int vdu			;

		; Cursor off

		mov ch,20h		; Start line
		mov cl,20h		; End line

		mov ah,1		; Set size
		int vdu			; Call it

		; Set the mode

		mov ax,0B800h		; Colour address
		mov cs:Wind_Seg,ax	; Save it

		; Draw the screen

Another_Game:	call Draw_Screen	; Clear it, etc

		mov b cs:Esc_Flag,0	; No escape yet

		call Init_Score		; 0

		call Init_T_Keys	; Clear the keys

		call Play_Tet_Sheet	; Play one game

NG_Lp:		call far ptr Far_Next_Frame ; Delay a bit

		call TRand_32		; Scramble the number

		call Scan_TKeys		; Esc or Drop ?

		cmp al,tkDrop		; Yes ?
		je Another_Game		; Loop

		cmp al,tkEsc		; Esc ?
		jne NG_Lp		; No, loop

PG_Esc:		retf			; Done

; Setup the display

Draw_Screen:	mov ax,cs:Wind_Seg	; Point at the display
		mov es,ax		;

		mov di,0		; Top-Left

		mov ax,01B0h		; Blue
		mov cx,80*2*25		; Size

		rep stosw		; Set it

		; Draw a box for the game

		mov di,19*2		; Top-left

		mov w es:[di],00BDAh	;

		mov cx,fWidth*2		; Width
		mov bx,2		; Step
		mov ax,0BC4h		; Char

		call DS_Lp		; Draw it

		mov w es:[di],0BBFh	; Top-right

		mov cx,fDepth		; Depth
		mov bx,2*Wide		; Step
		mov ax,0BB3h		; Char

		call DS_Lp		; Draw it

		mov w es:[di],00BD9h	; Bottom-right

		mov cx,fWidth*2		; Width
		mov bx,-2		; Step
		mov ax,0BC4h		; Char

		call DS_Lp		; Draw it

		mov w es:[di],0BC0h	; Bottom-left

		mov cx,fDepth		; Depth
		mov bx,-2*Wide		; Step
		mov ax,0BB3h		; Char

		call DS_Lp		; Draw it

		; Now, the score

		mov di,0		; Top-left

		mov w es:[di],00BDAh	;

		mov cx,14		; Width
		mov bx,2		; Step
		mov ax,0BC4h		; Char

		call DS_Lp		; Draw it

		mov w es:[di],0BBFh	; Top-right

		mov cx,1		; Depth
		mov bx,2*Wide		; Step
		mov ax,0BB3h		; Char

		call DS_Lp		; Draw it

		mov w es:[di],00BD9h	; Bottom-right

		mov cx,14		; Width
		mov bx,-2		; Step
		mov ax,0BC4h		; Char

		call DS_Lp		; Draw it

		mov w es:[di],0BC0h	; Bottom-left

		mov cx,1		; Depth
		mov bx,-2*Wide		; Step
		mov ax,0BB3h		; Char

		call DS_Lp		; Draw it

		; Now the "score = " string

		push cs			; Src in cs
		pop ds			;

		mov cx,14		; This long

		mov si,o sTScore	; Score string
		mov di,162		; Posn

		rep movsw		; Copy it

		ret			; Done

	; Write a line

DS_Lp:		add di,bx		; Step
		mov es:[di],ax		; Write it

		loop DS_Lp		; Loop

		add di,bx		; Skip last

		ret			; Done

	; The score name

sTScore		db "S",0Eh,"c",0Eh,"o",0Eh,"r",0Eh
		db "e",0Eh," ",0Eh,"=",0Eh," ",0Eh
		db " ",0Eh," ",0Eh," ",0Eh," ",0Eh
		db " ",0Eh,"0",0Eh

; Reset the keys

Init_T_Keys:	mov si,o Key_Vars	; Clear the vars
		mov cx,6*2		; 12 bytes
		mov al,0		; 00

		jmp CM_Fill		; Clear them to 0

; Get a key

Scan_TKeys:	mov si,o Key_Vars	; Point at the work-space

		mov al,0		; Nothing yet

		call far ptr Far_Test_Left ; Left ?
		call STK_Scan		   ; Do it

		call far ptr Far_Test_Right;
		call STK_Scan		   ;	

		call far ptr Far_Test_RotL ;
		call STK_Scan		   ;

		call far ptr Far_Test_RotR ;
		call STK_Scan		   ;

		call far ptr Far_Test_Drop ;
		call STK_Scan		   ;

		call far ptr Far_Test_Esc  ;
		
	; scan ?

STK_Scan:	jne STKS_P		; Pressed, skip

		; Not pressed, kill the repeat

		mov w cs:[si],0		; Kill this key
		lea si,2[si]		; Skip

		ret			; Done

	; Pressed, should we repeat it ?

STKS_P:		test b cs:1[si],-1	; Pressed ?
		jz STKS_First		; No, return it

		mov ah,cs:[si]		; Get the count

		cmp ah,kRep_Cnt		; Send it yet ?
		je STKSP_1		; Yes, skip

		inc ah			; Step it
		mov cs:[si],ah		; Save

STKS_X:		lea si,2[si]		; Skip
		ret			; Done

	; First press

STKS_First:	mov b cs:[si],-15	; Pause longer

		cmp al,0		; Any others ?
		jnz STKS_X		; Yes, skip

		jmp s STKS_Hit		; Hit it

	; Try to press it

STKSP_1:	cmp al,0		; Any key yet ?
		jne STKS_X		; Yes, exit

		; Clear the repeat

		mov b cs:[si],0		; Delay = 0

		; Return this key

STKS_Hit:	mov ax,si		; Point at it
		sub ax,o Key_Vars	; Offset
		shr ax,1		; / 2

		inc al			; 1 .. 6

		; Setup repeat rate

		mov b cs:1[si],01h	; Set pressed

		jmp s STKS_X		; Done

; Play one sheet

Play_Tet_Sheet:	mov w cs:Delay,Initial_Speed; Set speed

		call Clear_Map		; Clear it

PS_Lp:		call Get_Shape		; Get the next shape

		call Test_Current	; Does it fit ?
		jc PS_End		; No, done

		mov b cs:Fall,0		; Not gone far yet

PS_Lp1:		inc b cs:Fall		; Fallen once more

		call Player_Move	; Let the player try it

		test b cs:Esc_Flag,1	; Exit requested ?
		jne PS_End		; Yes, exit

		call Show_Score		; Display it

		call Move_Down		; Step the shape down
		jnc PS_Lp1		; Ok

		; Hit bottom

		call Compress		; Remove any full rows

		mov bx,Step_One		; Step the delay this much if dropped

		test b cs:Drop_F,0FFh	; Dropped ?
		jnz PS_Drop		; Yes				

		; Did it fall far enough ?

		mov al,cs:Fall		; How far has it dropped ?

		cmp al,3		; At least three
		jc PS_Drop		; No, don't penalise

		; So, the player didn't drop it

		mov bx,Step_Two		; Worse if not dropped

PS_Drop:	add cs:Delay,bx		; Step the delay

		mov ax,cs:Delay		; Fastest possible ?
		cmp ax,Base_Speed	;
		ja PS_Lp		; No, skip

		mov w cs:Delay,Loop_Speed ; Clip to this speed

		jmp PS_Lp		; Loop

PS_End:		ret			; Done

; Move the shape down one line if possible, else draw and CF

Move_Down:	call Pause_Test		; Wait if required

		inc b cs:sY		; Step down

		call Test_Current	; OK ?
		jnc MD_Done		; Yes, continue

		dec b cs:sY		; Step back

		call Draw_Shape		; Place it here

		stc			; Fail
MD_Done:	ret			;

; Allow the player to move the block for a while

Player_Move:	mov cx,cs:Delay		; Get the time

		shr cx,1		; / 32
		shr cx,1		;
		shr cx,1		;
		shr cx,1		;
		shr cx,1		;

PM_Lp:		push cx			; Save the timer

		call PM_DoIt		; Move the block

		call Draw_Shape		; Place the object

		call Show_Map		; Put it on the screen

		call Undraw_Shape	; Remove it

		call far ptr Far_Next_Frame ; Wait 1/70'th sec

		pop cx			; Restore

		; Can we drop it yet ?

		mov al,cs:Fall		; How far has it dropped

		cmp al,3		; At least two
		jc PM_NF		; Not yet

		; Should we drop it ?

		test b cs:Drop_F,0FFh	; Dropped ?
		jnz PM_Dropped		; Yes

		; Not dropped yet

PM_NF:		mov b cs:Drop_F,0	; Clear it

		; Loop

		loop PM_Lp		; Loop

		jmp PM_Exit		; Done

PM_Dropped:	mov ax,0001h		; Score + 1
		call Inc_Score		;

PM_Exit:	ret			; Done

; If key pressed, then move block

PM_DoIt:	call TRand_32		; Scramble it

		call Scan_TKeys		; Scan the keys

		mov bh,0		; No angle change

		mov bl,0		; No L/R

		cmp al,tkEsc		; Esc ?
		jz PM_Esc		; Yes, skip

		; L/R

		cmp al,tkLeft		; Left ?
		jnz PMDI_L		; No, skip

		dec bl			; Left

PMDI_L:		cmp al,tkRight		; Right ?
		jnz PMDI_R		; No, skip

		inc bl			; Right

		; Rotates

PMDI_R:		cmp al,tkRotL		; Rot Left ?
		jnz PMDI_RL		; No, skip

		dec bh			; Left

PMDI_RL:	cmp al,tkRotR		; Rot Right ?
		jnz PMDI_RR		; No, skip

		inc bh			; Right

		; Drop ?

PMDI_RR:	cmp al,tkDrop		; Drop ?
		jnz PMDI_D		; No, skip

		mov b cs:Drop_F,1	; Assert drop

PMDI_D:		cmp bx,0		; Moved ?
		jnz PM_DI_TryIt		; Try it

		clc			; Ok
		ret			; Done

PM_Esc:		mov b cs:Esc_Flag,1	; Exit requested
		ret			; Done

; Now move it if the new position and angle are OK

PM_DI_TryIt:	mov dl,cs:sX		; Calc its X
		add dl,bl		;

		mov dh,cs:sAngle	; Calc its angle
		add dh,bh		;
		and dh,03h		; Mask

		call Test_Shape_Fit	; Will it fit in this place ?
		jc PM_DI_Done		; No, ignore this

		mov cs:sX,dl		; Move the object
		mov cs:sAngle,dh	;		

PM_DI_Done:	clc			; Done
		ret			;

; See if there are any complete lines and remove them if so

Compress:	mov di,o MAP_Start+((fDepth-1)*(fWidth+2)) ; Point at last line

		mov cx,fDepth		; Number of lines to try

Comp_Lp:	push cx			; Save count

		call Comp_Test		; Is this line full ?
		jnc Comp_1		; No, skip

		call Comp_Crush		; Remove this line

		push di			; Save ptr

		call Show_MAP		; Show it

		pop di			; Restore

Comp_Next:	pop cx			; Do next
		loop Comp_Lp		;

		ret			; Done

Comp_1:		sub di,fWidth+2		; Step back up
		jmp Comp_Next		; Do the next line

; Test if the line (di) is full

Comp_Test:	mov cx,fWidth		; This many wide

		mov bx,di		; Get ptr

CompT_Lp:	test b cs:[bx],0FFh	; Full ?
		jz CompT_Fail		; No, fail

		inc bx			; Step ptr

		loop CompT_Lp		; Loop

		stc			; Full
		ret			;

CompT_Fail:	clc			; Not full
		ret			;

; Remove the line at [di]

Comp_Crush:	mov bx,di		; Get ptr

CompC_Lp:	cmp bx,o MAP_Start	; Top line ?
		je CompC_Clear		; Yes, clear line

		mov cx,fWidth		; Number to copy

CompC_Lp1:	mov al,cs:0-(fWidth+2)[bx]; Read byte
		mov cs:[bx],al		  ; Move it

		inc bx			; Step across

		loop CompC_Lp1		; Loop across line

		sub bx,(fWidth+2)+fwidth; Step up one line

		jmp CompC_Lp		; Do next line

CompC_Clear:	mov cx,fWidth		; Clear this line

CompCC_Lp:	mov b cs:[bx],0		; Clear cell
		inc bx			;

		loop CompCC_Lp		; Loop

		mov ax,0100h		; 10 for crush
		call Inc_Score		;

		ret			; Done

; Is the shape clear ?

Test_Current:	mov dl,cs:sX		; X
		mov dh,cs:sAngle	; Angle

		jmp Test_Shape_Fit	; See if OK

; Test if shape (cShape), at X (dl), Y (sY) and Angle (dh)
; Overlaps any other blocks (CF Set if so)

Test_Shape_Fit:	mov al,cs:sShape	; Shape
		mov ah,dh		; Angle

		call Get_Shape_Ptr	; SI := Point at the shape's defn

		mov al,dl		; X
		mov ah,cs:sY		; Y

		call Get_Shape_Addr	; DI := Point at MAP

		mov bx,cs:1[si]		; First cell

		test b cs:[di+bx],0FFh	; Cell used ?
		jnz TSF_Fail		; Yes

		mov bx,cs:3[si]		; Next cell

		test b cs:[di+bx],0FFh	; Cell used ?
		jnz TSF_Fail		; Yes

		mov bx,cs:5[si]		; Next cell

		test b cs:[di+bx],0FFh	; Cell used ?
		jnz TSF_Fail		; Yes

		mov bx,cs:7[si]		; Last cell

		test b cs:[di+bx],0FFh	; Cell used ?
		jnz TSF_Fail		; Yes

		clc			; No overlap
		ret			;

TSF_Fail:	stc			; Overlap
		ret			;

; Put the shape into the map

Draw_Shape:	mov al,cs:sShape	; Shape
		mov ah,cs:sAngle	; Angle

		call Get_Shape_Ptr	; SI := Point at the shape's defn

		mov al,cs:sX		; X
		mov ah,cs:sY		; Y

		call Get_Shape_Addr	; DI := Point at MAP

		mov al,cs:[si]		; Get colour

		jmp DS_Draw		; Show it

; Remove the shape from the map

Undraw_Shape:	mov al,cs:sShape	; Shape
		mov ah,cs:sAngle	; Angle

		call Get_Shape_Ptr	; SI := Point at the shape's defn

		mov al,cs:sX		; X
		mov ah,cs:sY		; Y

		call Get_Shape_Addr	; DI := Point at MAP

		mov al,0		; Blank

DS_Draw:	mov bx,cs:1[si]		; First cell
		mov cs:[di+bx],al	; Write the cell

		mov bx,cs:3[si]		; First cell
		mov cs:[di+bx],al	; Write the cell

		mov bx,cs:5[si]		; First cell
		mov cs:[di+bx],al	; Write the cell

		mov bx,cs:7[si]		; First cell
		mov cs:[di+bx],al	; Write the cell

		clc			; No overlap
		ret			;

; Given al = shape, ah = angle : return si = ptr to defn of shape

Get_Shape_Ptr:	push ax			; Save
		push bx			;

		add al,al		; Shape * 4
		add al,al		;
		add al,ah		; Shape * 4 + Angle

		mov ah,0		; 16 bit

		mov bx,ax		; Save

		add ax,ax		; * 2
		add ax,ax		; * 4
		add ax,ax		; * 8

		add ax,bx		; * 9

		mov si,o Defn_Table	; Definition of shapes
		add si,ax		; Add offset

		pop bx			;
		pop ax			;
		ret			; Done

; Given al=X, ah=Y return ptr to MAP

Get_Shape_Addr:	push bx			; Save reg
		push ax			;

		mov bl,al		; X
		mov al,fWidth+2		; Width of MAP
		imul ah			; Y*Width

		xchg ax,bx		; Swap them

		cbw			; x := 16 bit

		add ax,bx		; ax := x+(y*width)

		mov di,o MAP_Start	; Base
		add di,ax		; Full address

		pop ax			; Restore
		pop bx			;

		ret			; Done

; Get a number from 0 to 6 randomly

Get_Shape:	call TRand_32		; Get us a random number

		and ax,07h		; Mask it
		jz Get_Shape		; 1..7 only

		dec al			; 0..6

		mov cs:sShape,al	; Store it

		mov b cs:sX,fWidth/2	; X
		mov b cs:sY,1		; Y
		mov b cs:sAngle,0	; Angle

		mov b cs:Drop_F,0	; Clear dropped flag

		ret			; Done

; Clear the playing area, should use STOSB (if I could remember the regs !)

Clear_Map:	mov si,o MAP		; Fill with 0

		mov cx,fDepth+1		; Length of data

CM_Lp:		push cx			; Save count

		mov b cs:[si],0FFh	; Guard
		inc si			;

		mov al,0		; Play area
		mov cx,fWidth		; Length

		call CM_Fill		; Fill it

		mov b cs:[si],0FFh	; Guard
		inc si			;

		pop cx			; Loop
		loop CM_Lp		;

		mov al,0FFh		; Guard
		mov cx,fWidth+2		; Length

CM_Fill:	mov cs:[si],al		; Fill it
		inc si			;

		loop CM_Fill		; Loop

		ret			;
		
; Write data from (CS:SI) to screen

Show_Map:	push ds			; Save segment registers

		mov ds,cs:Wind_Seg	; Buffer segment

		mov si,o MAP_Start	; Point at the map

		mov di,20*2+Wide*2	; Point at the first usable char

		mov al," "		; Space char

		mov bx,fDepth		; 20 lines

SM_Lp1:		mov cx,fWidth		; 10 cells per line

SM_Lp2:		mov ah,cs:[si]		; Read source
		inc si			; Step

		add ah,ah		; Align it
		add ah,ah		;
		add ah,ah		;
		add ah,ah		;

		and ah,70h		; Mask
		or ah,08h		; Bright

		mov [di],ax		; Write char + attr
		mov 2[di],ax		; Twice

		add di,4		;
		
		loop SM_Lp2		; Loop X

		add si,2		; Skip Guard bytes

		add di,Wide*2-fWidth*4	; Step to next line

		dec bx			; Loop Y
		jnz SM_Lp1		;		

		pop ds			; Restore

		ret			; Done

; Add ax to score (done in BCD to make the display routines easier)

Inc_Score:	add al,cs:TScore	; Add score
		daa			;
		mov cs:TScore,al	;

		mov al,ah		;

		adc al,cs:TScore+1	;
		daa			;
		mov cs:TScore+1,al	;

		mov al,0		;
		adc al,cs:TScore+2	;
		daa			;
		mov cs:TScore+2,al	;

		ret			; Done

; Zero score

Init_Score:	mov al,0		; Clear score

		mov cs:TScore,al	;
		mov cs:TScore+1,al	;
		mov cs:TScore+2,al	;

		ret			; Done

; Put score onto display

Show_Score:	mov es,cs:Wind_Seg	; Segment
		mov di,160+9*2		; Create string

		mov ah," "		; Zero suppress

		mov cx,3		; 6 digit (3 bytes)

		mov si,o TScore+2	; MSB

SS_Lp:		mov al,cs:[si]		; MSBs
		call SS_Byte		;

		dec si			;

		loop SS_Lp		; Loop

		cmp ah," "		; 0 ?
		jne SS_Done		; No, exit

		mov b es:-2[di],"0"	; At least 0

SS_Done:	ret		

SS_Byte:	mov bl,al		; Save byte
		
		shr al,1		; Align
		shr al,1		;
		shr al,1		;
		shr al,1		;

		call SSB_Digit		; Do one digit

		mov al,bl		; Restore

SSB_Digit:	and al,0Fh		; Mask

		jz SSBD_0		; "0"

		mov ah,"0"		; Not leading now

		or al,ah		; Make ASCII

		mov es:[di],al		; Store
		inc di			; Step
		inc di			;

		ret			; Done

SSBD_0:		mov es:[di],ah		; " " or "0"
		inc di			;
		inc di			;

		ret			; Done

; Pause 

Pause_Test:	call Test_PK		; Pause ?
		jz PT_X			; No, skip

PT1:		call Test_PK_NF		; Until released
		jnz PT1			;

PT2:		call Test_PK_NF		; Until pressed
		jz PT2			;

PT3:		call Test_PK_NF		; Until released again
		jnz PT3			;

PT_X:		ret			;

Test_PK_NF:	call far ptr Far_Next_Frame ; Wait a bit

Test_PK:	call far ptr Far_Test_Pause ; Try it
		ret			    ; Done

; Random numbers

TRand_32:	push cx			;
		push bx			;

		call TRand_Inner	; Get next

		xor al,dh		; Scramble it a bit
		xor ah,dl		;

		pop bx			; Restore
		pop cx			;

		ret			; Done

TRand_Inner:	mov ax,cs:TSeed		; This is similar to Turbo Pascal
		mov bx,cs:TSeed+2	;
		mov cx,ax		;

		mul w cs:TNasty	 	;

		shl cx,1		;
		shl cx,1		;
		shl cx,1		;

		add ch,cl		;
		add dx,cx		;
		add dx,bx		;

		shl bx,1		;
		shl bx,1		;

		add dx,bx		;
		add dh,bl		;

		mov cl,5		;
		shl bx,cl		;

		add dh,bl		;

		add ax,1		;
		adc dx,0		;

		mov cs:TSeed,ax		;
		mov cs:TSeed+2,dx	;

		ret			;

; Data for shapes

xs		equ 1			; One cell in X
ys		equ (fWidth+2)		; One cell in Y

Defn_Table	db 4			; 0 (bar)
		dw -xs,0,xs,2*xs	;
		db 4			;
		dw -ys,0,ys,2*ys	;
		db 4			;
		dw -2*xs,-xs,0,xs	;
		db 4			;
		dw -2*ys,-ys,0,ys	;

		db 1			; 1 (block)
		dw 0,xs,ys,ys+xs	;
		db 1			;
		dw 0,-xs,ys,ys-xs	;
		db 1			;
		dw 0,-xs,-ys,-ys-xs	;
		db 1			;
		dw 0,xs,-ys,xs-ys	;

		db 6			; 2 (point)
		dw 0,-xs,xs,ys		;
		db 6			;
		dw 0,-xs,-ys,ys		;
		db 6			;
		dw 0,-xs,xs,-ys		;
		db 6			;
		dw 0,xs,-ys,ys		;

		db 7			; 3 (L)
		dw -xs,0,xs,xs+ys	;
		db 7			;
		dw -ys,0,ys,ys-xs	;
		db 7			;
		dw -xs,0,xs,-xs-ys	;
		db 7			;
		dw -ys,xs-ys,0,ys	;

		db 3			; 4 (RL)
		dw -xs,0,xs,xs-ys	;
		db 3			;
		dw -ys,0,ys,ys+xs	;
		db 3			;
		dw -xs,0,xs,ys-xs	;
		db 3			;
		dw -ys,-xs-ys,0,ys	;

		db 2			; 5 (xx
		dw -xs,0,ys,ys+xs	;     xx)
		db 2			;
		dw -xs,0,-ys,ys-xs	;
		db 2			;
		dw 0,xs,-xs-ys,-ys	;
		db 2			;
		dw ys,0,xs,xs-ys	;

		db 5			; 6 ( xx
		dw xs,0,ys,ys-xs	;    xx )
		db 5			;
		dw -xs,0,-ys-xs,ys	;
		db 5			;
		dw 0,-xs,xs-ys,-ys	;
		db 5			;
		dw -ys,0,xs,xs+ys	;

TNasty		dw 33797		; Nasty number
TSeed		dw 2781,42		; Random number seed

TCurr_Key	db 0			; Current keycode
TKey_Waiting	db 0			; Key flag

Wind_Seg	dw ?			; The segment addr of the window buff

Delay		dw ?

sX		db ?			; X of shape
sY		db ?			; Y of shape
sShape		db ?			; Current shape
sAngle		db ?			; Angle

Drop_F		db ?			; Dropped flag

Esc_Flag	db ?			; Escape flag

Fall		db ?			; How far

Key_Vars	dw 6 dup (?)		; Keys

TScore		db 3 dup (?)		; 6 Digits

MAP		db (fWidth+2)+1 dup (?)	; Guard
MAP_Start	db (fWidth+2)*(fDepth+2) dup (?); Play area
